#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <signal.h>

#include <winsock2.h>
#include <windows.h>


#define PRIVATE static
#define PUBLIC

#define CHAR_EOS ((char)'\0')
#define CHAR_NL ((char)'\n')


typedef enum
{
  E_FALSE = 0,
  E_TRUE = 1
} TE_BOOL;

typedef struct
{
  SOCKET o_socket;
  SOCKADDR_IN o_sockaddr;
  char *pc_rcon_password;
} TS_NET_INFO;


volatile TE_BOOL gb_quit = E_FALSE;


PRIVATE TE_BOOL r_net_init( TS_NET_INFO *pts_net_info,
                            char *pc_address,
                            char *pc_rcon_password )
{
  WORD o_version_requested;
  WSADATA o_wsa_data;
  int i_result;
  char *pc_ptr;
 
  o_version_requested = MAKEWORD( 2, 0 );
 
  i_result = WSAStartup( o_version_requested, &o_wsa_data );
  if ( i_result != 0 )
  {
    return E_FALSE;
  }

  if ( LOBYTE( o_wsa_data.wVersion ) != 2 ||
        HIBYTE( o_wsa_data.wVersion ) != 0 )
  {
    WSACleanup();
    return E_FALSE;
  }

  pts_net_info->o_socket = socket( PF_INET, SOCK_DGRAM, IPPROTO_UDP );
  if ( pts_net_info->o_socket == INVALID_SOCKET )
  {
    WSACleanup();
    return E_FALSE;
  }

  pc_ptr = strchr( pc_address, ':' );
  if ( pc_ptr == NULL )
  {
    WSACleanup();
    return E_FALSE;
  }
  *pc_ptr = CHAR_EOS;

  memset( &(pts_net_info->o_sockaddr), 0, sizeof(pts_net_info->o_sockaddr) );
  pts_net_info->o_sockaddr.sin_family = AF_INET;
  pts_net_info->o_sockaddr.sin_addr.s_addr = inet_addr( pc_address );
  pts_net_info->o_sockaddr.sin_port = htons( (short)atoi( pc_ptr + 1 ) );

  pts_net_info->pc_rcon_password = pc_rcon_password;

  return E_TRUE;
}

PRIVATE void r_net_shutdown( SOCKET o_socket )
{
  int i_result;

  i_result = shutdown( o_socket, SD_BOTH );
  if ( i_result == SOCKET_ERROR )
  {
    WSACleanup();
    return;
  }

  i_result = closesocket( o_socket );
  if ( i_result == SOCKET_ERROR )
  {
    WSACleanup();
    return;
  }

  i_result = WSACleanup();
  if ( i_result == SOCKET_ERROR )
  {
    return;
  }

  return;
}


#define D_PACKET_SIZE 1500

PRIVATE size_t r_net_receive_print( TS_NET_INFO *pts_net_info )
{
  char ac_receive_data[D_PACKET_SIZE + 1];
  struct sockaddr_in o_rcv_address;
  int i_rcv_len;
  int i_receive_len;
  fd_set o_rfds;
  struct timeval o_tv;
  int i_result;

  // Watch socket to see if it has input
  FD_ZERO( &o_rfds );
  FD_SET( pts_net_info->o_socket, &o_rfds );

  // No wait timeout; Return immediately
  o_tv.tv_sec = 0;
  o_tv.tv_usec = 0;

  // Wait for data to become available
  i_result = select( pts_net_info->o_socket + 1, &o_rfds, NULL, NULL, &o_tv );

  if (
       ( i_result > 0 ) &&
       ( FD_ISSET( pts_net_info->o_socket, &o_rfds ) )
     )
  {
    // Wait for reply
    i_rcv_len = sizeof( o_rcv_address );
    i_receive_len = recvfrom( pts_net_info->o_socket,
                              ac_receive_data,
                              D_PACKET_SIZE,
                              0,
                              (struct sockaddr *)&o_rcv_address,
                              &i_rcv_len );
    ac_receive_data[D_PACKET_SIZE] = CHAR_EOS;
    if (
         ( i_receive_len == -1 ) &&
         ( errno != EAGAIN )
       )
    {
      // Disabled because this happens for the windows version when the server isn't runnning
      // ERROR
      i_receive_len = 0;
    }
    else
    {
      if (
           ( i_receive_len == 0 ) ||
           ( errno == EAGAIN )
         )
      {
        // No data available
        i_receive_len = 0;
        errno = 0;
      }
      else
      {
        // Data received
        if ( strncmp( ac_receive_data, "\xFF\xFF\xFF\xFFprint\n", 10 ) == 0 )
        {
          ac_receive_data[i_receive_len] = CHAR_EOS;
          i_receive_len -= 10;
          if (
               ( i_receive_len > 0 ) &&
               ( o_rcv_address.sin_addr.s_addr == pts_net_info->o_sockaddr.sin_addr.s_addr ) &&
               ( o_rcv_address.sin_port == pts_net_info->o_sockaddr.sin_port )
             )
          {
            printf( "%s", &ac_receive_data[10] );
          }
        }
      }
    }
  }
  else
  {
    // Select timeout if i_result == 0 -> No data available within timeout
    i_receive_len = 0;
  }

  return i_receive_len;
}


#define D_COMMAND_SIZE 1024

PRIVATE void r_net_send_rcon( TS_NET_INFO *pts_net_info, char *pc_command, ... )
{
  char ac_command[D_COMMAND_SIZE + 1];
  va_list o_varargs;
  int i_length;
  int i_result;

  i_length = _snprintf( ac_command,
                        D_COMMAND_SIZE,
                        "\xFF\xFF\xFF\xFFrcon \"%s\" ",
                        pts_net_info->pc_rcon_password );
  if ( i_length < 0 )
  {
    return;
  }

  va_start( o_varargs, pc_command );
  i_result = _vsnprintf( &ac_command[i_length],
                         D_COMMAND_SIZE - i_length,
                         pc_command,
                         o_varargs );
  va_end( o_varargs );
  if ( i_result < 0 )
  {
    return;
  }

  i_length += i_result;

  i_result = sendto( pts_net_info->o_socket,
                     ac_command,
                     i_length + 1,
                     0,
                     (struct sockaddr *)&(pts_net_info->o_sockaddr),
                     sizeof(pts_net_info->o_sockaddr) );
  if ( i_result == SOCKET_ERROR )
  {
    return;
  }

  return;
}


void r_signal( int i_sig )
{
  signal( i_sig, r_signal );

  gb_quit = E_TRUE;
}

#define D_MAX_LINE_LEN 1024

int main( int argc, char *argv[] )
{
  TS_NET_INFO ts_net_info;
  TE_BOOL b_result;
  TE_BOOL b_loop;
  FILE *po_in_file;
  char ac_line[D_MAX_LINE_LEN + 1];
  size_t o_len;
  char *pc_result;
  time_t o_time;
  int i_len;

  if ( argc != 5 )
  {
    exit( EXIT_FAILURE );
  }

  signal( SIGABRT, r_signal );
  signal( SIGINT, r_signal );
  signal( SIGTERM, r_signal );

  if ( strcmp( argv[3], "loop" ) == 0 )
  {
    b_loop = E_TRUE;
  }
  else
  {
    b_loop = E_FALSE;
  }

  b_result = r_net_init( &ts_net_info,
                         argv[1],
                         argv[2] );
  if ( b_result == E_FALSE )
  {
    exit( EXIT_FAILURE );
  }

  gb_quit = E_FALSE;
  do
  {
    po_in_file = fopen( argv[4], "rt" );
    if ( po_in_file != NULL )
    {
      pc_result = fgets( ac_line, D_MAX_LINE_LEN + 1, po_in_file );
      while (
              ( pc_result != NULL ) &&
              ( gb_quit == E_FALSE )
            )
      {
        o_len = strlen( ac_line );
        if (
             ( o_len > 0 ) &&
             ( ac_line[o_len - 1] == CHAR_NL )
           )
        {
          ac_line[o_len - 1] = CHAR_EOS;
          o_len--;
        }

        if ( strncmp( ac_line, "rcon ", 5 ) == 0 )
        {
          r_net_send_rcon( &ts_net_info, &ac_line[5] );
        }
        else if ( strncmp( ac_line, "echo ", 5 ) == 0 )
        {
          printf( "%s\n", &ac_line[5] );
        }
        else if ( strncmp( ac_line, "wait ", 5 ) == 0 )
        {
          // Get current time
          o_time = time( NULL );
          // Caclulate time when to continue
          o_time += atol( &ac_line[5] );

          // Sleep until next query
          while (
                  ( gb_quit == E_FALSE ) &&
                  ( o_time > time( NULL ) )
                )
          {
            // Receive print messages from server and display them
            do
            {
              // Receive data (no wait)
              i_len = r_net_receive_print( &ts_net_info );
            }
            while ( i_len > 0 );

            // Sleep 0.5 seconds
            Sleep( 500 );
          }
        }

        // Receive print messages form server and display them
        do
        {
          // Receive data (no wait)
          i_len = r_net_receive_print( &ts_net_info );
        }
        while ( i_len > 0 );

        pc_result = fgets( ac_line, D_MAX_LINE_LEN + 1, po_in_file );
      }

      fclose( po_in_file );
    }
  }
  while (
          ( gb_quit == E_FALSE ) &&
          ( b_loop == E_TRUE )
        );

  r_net_shutdown( ts_net_info.o_socket );

  return 0;
}
